@@ -1,18 +1,20 @@
 import netsblox as _netsblox
+from netsblox import sound as _Sound
 
 import randomname as _randomname
 import threading as _threading
 import requests as _requests
+import difflib as _difflib
 import inspect as _inspect
 import base64 as _base64
+import numpy as _np
 import json as _json
 import sys as _sys
 import io as _io
 import os as _os
-
 from PIL import Image as _Image, ImageTk as _ImageTk
 
-from typing import Tuple, List, Any, Optional
+from typing import Tuple, List, Any, Optional, Dict
 
 _NETSBLOX_PY_PATH = _os.path.dirname(_netsblox.__file__)
 
@@ -65,10 +67,16 @@ def nothrow(f):
             return msg
     return wrapped
 
+def get_antialias_mode():
+    # for god knows why, the pillow devs decided to break everything that used Image.ANTIALIAS when they added different resampling modes
+    # so we need this wrapper to get whichever one works on the installed version of the package at runtime... eww...
+    res = getattr(_Image, 'ANTIALIAS', None)
+    return res if res is not None else _Image.Resampling.BICUBIC
+
 def scale_image(img: _Image.Image, scale: float = 1) -> _Image.Image:
     if scale == 1: return img
     new_size = tuple(round(v * scale) for v in img.size)
-    return img.resize(new_size, resample = _Image.ANTIALIAS)
+    return img.resize(new_size, resample = get_antialias_mode())
 
 _img_lock = _threading.Lock()
 _img_cache = {}
@@ -81,6 +89,8 @@ def load_image(uri: str) -> _Image.Image:
             return _img_cache[uri]
         elif protocol == 'netsblox':
             img = _Image.open(f'{_NETSBLOX_PY_PATH}/{uri[11:]}')
+        elif protocol == 'base64':
+            return decode_image(uri[9:]) # return so we don't cache local data (also the uri is long)
         elif protocol in ['https', 'http']:
             res = _requests.get(uri)
             if res.status_code == 200:
@@ -119,11 +129,19 @@ def load_text(uri: str) -> str:
         _text_cache[uri] = txt
         return txt
 
-def generate_proj_id() -> str:
-    return f'py-{_randomname.get_name()}'
+def generate_project_id() -> str:
+    return f'_py-{_randomname.get_name()}'
 
 def small_json(obj):
-    return _json.dumps(obj, separators=(',', ':'))
+    def prep_value(obj):
+        if type(obj) in [list, tuple]:
+            return [prep_value(x) for x in obj]
+        if type(obj) is dict:
+            return { prep_value(k): prep_value(v) for k,v in obj.items() }
+        if type(obj) is _np.ndarray:
+            return obj.tolist()
+        return obj
+    return _json.dumps(prep_value(obj), separators=(',', ':'))
 
 def prep_send(val):
     if val is None:
@@ -157,13 +175,15 @@ def get_location() -> Tuple[float, float]:
 
     Note that an internet connection is required for this to work.
     '''
-    res = _requests.post('https://reallyfreegeoip.org/json/', headers = { 'Content-Type': 'application/json' })
-
-    if res.status_code == 200:
-        parsed = _json.loads(res.text)
-        return parsed['latitude'], parsed['longitude']
-    else:
-        raise Exception(f'Failed to get location: {res.status_code}\n{res.text}')
+    try:
+        res = _requests.get('https://reallyfreegeoip.org/json/', headers = { 'Content-Type': 'application/json' })
+        if res.status_code == 200:
+            parsed = _json.loads(res.text)
+            return parsed['latitude'], parsed['longitude']
+        else:
+            raise Exception(f'Failed to get location: {res.status_code}\n{res.text}')
+    except:
+        return (0.0, 0.0) # instead of hard failing, just give an obvious invalid value
 
 def encode_image(img: _Image.Image) -> str:
     res = _io.BytesIO()
@@ -173,6 +193,12 @@ def decode_image(img: str) -> _Image.Image:
     raw = _base64.decodebytes(img.encode('ascii'))
     return _Image.open(_io.BytesIO(raw))
 
+def encode_sound(snd: _Sound.Sound) -> str:
+    return _base64.b64encode(getattr(snd, '_Sound__raw', snd)).decode('ascii')
+def decode_sound(snd: str) -> _Sound.Sound:
+    raw = _base64.decodebytes(snd.encode('ascii'))
+    return _Sound.Sound(raw)
+
 def paginate_str(text: str, max_len: int) -> List[str]:
     res = ['']
     pos = 0
@@ -191,9 +217,10 @@ def paginate_str(text: str, max_len: int) -> List[str]:
         word = text[pos:end]
         pos = end
 
+        # if len(res[-1]) == 0: res[-1] = word
         if len(res[-1]) + len(word) <= max_len: res[-1] += word
         else: res.append(word)
-    return res
+    return [x.strip() for x in res]
 
 class PointerSet:
     def __init__(self):
@@ -217,6 +244,30 @@ class PointerSet:
             return True
         return False
 
+def unified_diff(before: str, after: str, *, n: int = 3) -> str:
+    before = [f'{x}\n' for x in before.splitlines()]
+    after = [f'{x}\n' for x in after.splitlines()]
+    return ''.join(_difflib.unified_diff(before, after, n = n))
+
+class Namespace:
+    def __init__(self, src: Dict[str, Any]):
+        self.__dict__['__Namespace_src'] = src
+
+    def __getattr__(self, name: str) -> Any:
+        src = self.__dict__['__Namespace_src']
+        if name not in src:
+            raise AttributeError(name)
+        return src[name]
+
+    def __setattr__(self, name: str, value: Any) -> None:
+        self.__dict__['__Namespace_src'][name] = value
+
+    def __delattr__(self, name: str) -> None:
+        src = self.__dict__['__Namespace_src']
+        if name not in src:
+            raise AttributeError(name)
+        del src[name]
+
 if __name__ == '__main__':
     import sys
     failures = [0]
@@ -230,11 +281,13 @@ if __name__ == '__main__':
     assert_eq(paginate_str('hello world', 20), ['hello world'])
     assert_eq(paginate_str('hello\nworld', 20), ['hello', 'world'])
     assert_eq(paginate_str('hello\nworld\n', 20), ['hello', 'world', ''])
-    assert_eq(paginate_str('hello world', 10), ['hello ', 'world'])
-    assert_eq(paginate_str('hello world this is a long message', 15), ['hello world ', 'this is a long ', 'message'])
-    assert_eq(paginate_str('hello world this is a long message', 16), ['hello world this', ' is a long ', 'message'])
-    assert_eq(paginate_str('hello world this is a long message', 17), ['hello world this ', 'is a long message'])
-    assert_eq(paginate_str('empty                     space', 10), ['empty     ', '          ', '      ', 'space'])
+    assert_eq(paginate_str('hello world', 10), ['hello', 'world'])
+    assert_eq(paginate_str('hello world this is a long message', 15), ['hello world', 'this is a long', 'message'])
+    assert_eq(paginate_str('hello world this is a long message', 16), ['hello world this', 'is a long', 'message'])
+    assert_eq(paginate_str('hello world this is a long message', 17), ['hello world this', 'is a long message'])
+    assert_eq(paginate_str('empty                     space', 10), ['empty', '', '', 'space'])
+    assert_eq(paginate_str('hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh', 10), ['hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh'])
+    assert_eq(paginate_str('h hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh', 10), ['h', 'hhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh'])
 
     ps = PointerSet()
     assert_eq(len(ps), 0)
@@ -273,3 +326,5 @@ if __name__ == '__main__':
         print(f'FAILED TESTS: {failures[0]}', file = sys.stderr)
         sys.exit(1)
     print(f'passed all {total[0]} tests')
+
+
