@@ -1,8 +1,16 @@
 import random
 import math
+import re
+import io
+import json
+import csv
+import itertools
+from PIL import Image
 
 from typing import Any, Union, Callable, Sequence
 
+import netsblox as _netsblox
+
 def _is_matrix(v: Any) -> bool:
     return isinstance(v, list) and len(v) > 0 and isinstance(list.__getitem__(v, 0), list)
 def _is_list(v: Any) -> bool:
@@ -23,6 +31,11 @@ def _list_unary_op(a: Any, op: Callable) -> 'List':
     if _is_list(a):
         return List(_list_unary_op(wrap(x), op) for x in a)
     return op(a)
+def _list_fold_op(a: Any, acc: Any, op: Callable) -> 'List':
+    assert is_wrapped(acc) and is_wrapped(a) and _is_list(a)
+    for i in range(len(a)):
+        acc = op(acc, wrap(list.__getitem__(a, i)))
+    return acc
 
 def _scalar_op(a: Any, b: Any, op: Callable) -> 'Float':
     a, b = float(a), float(b)
@@ -31,14 +44,23 @@ def _scalar_op(a: Any, b: Any, op: Callable) -> 'Float':
     except OverflowError:
         return Float(math.inf)
 
-def _parse_index(idx: Any) -> Union['List', int, str]:
+def _single_cmp(a: Any, b: Any) -> int:
+    base = lambda x, y: 1 if x > y else -1 if x < y else 0
+    try:
+        return base(float(a), float(b))
+    except:
+        return base(str(a), str(b))
+
+def _parse_index(idx: Any) -> Union['List', int, str, slice]:
     idx = wrap(idx)
-    if type(idx) is List: return idx
+    if type(idx) is List or type(idx) is slice: return idx
 
     try:
-        idx = +idx
+        x = +idx
+        if isinstance(x, int) or (isinstance(x, float) and not math.isnan(x)):
+            idx = x
     except:
-        return str(idx)
+        pass
 
     return idx if type(idx) is int else str(idx)
 
@@ -49,9 +71,26 @@ def _float_div(a: Any, b: Any) -> 'Float':
     return Float(math.inf if a > 0 else -math.inf)
 
 class Float(float):
+    def __new__(cls, v: Any):
+        try:
+            return float.__new__(cls, v)
+        except:
+            return float.__new__(cls, math.nan)
+
     def __eq__(self, other: Any) -> bool:
         if isinstance(other, str): return wrap(float(self) == float(other))
         return wrap(float(self) == other)
+    def __ne__(self, other: Any) -> bool:
+        return not (self == other)
+
+    def __gt__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) > 0)
+    def __ge__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) >= 0)
+    def __lt__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) < 0)
+    def __le__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) <= 0)
 
     def __str__(self) -> str:
         if math.isnan(self): return 'NaN'
@@ -131,16 +170,29 @@ class Str(str):
     def __getitem__(self, idx: Any) -> str:
         idx = _parse_index(idx)
         t = type(idx)
-        if t is List: return _list_unary_op(idx, lambda x: self[+x])
+        if t is slice: return wrap(str.__getitem__(self, idx))
+        elif t is List: return _list_unary_op(idx, lambda x: self[+x])
         elif t is str: return wrap('')
         elif t is int:
             if idx < 0 or idx >= len(self): return wrap('')
-            else: return str.__getitem__(self, idx)
+            else: return wrap(str.__getitem__(self, idx))
         else: assert False
 
     def __eq__(self, other: Any) -> bool:
         if isinstance(other, int) or isinstance(other, float): return float(self) == other
+        if isinstance(other, str): return str(self).lower() == str(other).lower()
         return str(self) == other
+    def __ne__(self, other: Any) -> bool:
+        return not (self == other)
+
+    def __gt__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) > 0)
+    def __ge__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) >= 0)
+    def __lt__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) < 0)
+    def __le__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) <= 0)
 
     def __add__(self, other: Any) -> Any:
         return Float(self) + other
@@ -193,6 +245,13 @@ class Str(str):
     def __floor__(self) -> 'Float':
         return math.floor(Float(self))
 
+    @property
+    def last(self) -> 'Str':
+        return wrap(str.__getitem__(self, -1) if len(self) != 0 else '')
+    @property
+    def rand(self) -> 'Str':
+        return wrap(str.__getitem__(self, random.randrange(len(self))) if len(self) != 0 else '')
+
 class List(list):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -203,10 +262,20 @@ class List(list):
     def __bool__(self) -> bool:
         return True
 
+    def __gt__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) > 0)
+    def __ge__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) >= 0)
+    def __lt__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) < 0)
+    def __le__(self, other: Any) -> Any:
+        return _list_binary_op(self, wrap(other), lambda a, b: _single_cmp(a, b) <= 0)
+
     def __delitem__(self, idx: Any) -> None:
         idx = _parse_index(idx)
         t = type(idx)
-        if t is List: pass
+        if t is slice: list.__delitem__(self, idx)
+        elif t is List: pass
         elif t is str: del self.__str_keys[idx]
         elif t is int:
             if idx < 0: del self.__str_keys[str(idx)]
@@ -216,7 +285,8 @@ class List(list):
     def __getitem__(self, idx: Any) -> Any:
         idx = _parse_index(idx)
         t = type(idx)
-        if t is List: return wrap('')
+        if t is slice: return wrap(list.__getitem__(self, idx))
+        elif t is List: return _list_unary_op(wrap(idx), lambda x: self[x])
         elif t is str: return wrap(self.__str_keys.get(idx, ''))
         elif t is int:
             if idx < 0: return wrap(self.__str_keys.get(str(idx), ''))
@@ -227,7 +297,8 @@ class List(list):
         value = wrap(value)
         idx = _parse_index(idx)
         t = type(idx)
-        if t is List: pass
+        if t is slice: list.__setitem__(self, idx, value)
+        elif t is List: pass
         elif t is str: self.__str_keys[idx] = value
         elif t is int:
             if idx < 0: self.__str_keys[str(idx)] = value
@@ -284,20 +355,115 @@ class List(list):
     def __floor__(self) -> 'List':
         return List(math.floor(x) for x in self)
 
-    def random(self) -> Any:
+    def __iadd__(self, other: Any) -> 'List':
+        return self + other
+
+    def append(self, value) -> None:
+        list.append(self, wrap(value))
+    def insert(self, idx, value) -> None:
+        list.insert(self, idx, wrap(value))
+    def insert_rand(self, value) -> None:
+        list.insert(self, random.randrange(len(self)) if len(self) != 0 else 0, wrap(value))
+
+    @property
+    def rand(self) -> Any:
         if len(self) == 0: return wrap('')
-        return wrap(random.choice(self))
+        return self[random.randrange(len(self))]
+    @rand.setter
+    def rand(self, value):
+        if len(self) == 0: return
+        self[random.randrange(len(self))] = wrap(value)
+
+    @property
     def last(self) -> Any:
         if len(self) == 0: return wrap('')
         return wrap(list.__getitem__(self, -1))
+    @last.setter
+    def last(self, value):
+        if len(self) == 0: return
+        list.__setitem__(self, -1, wrap(value))
+
+    def reshaped(self, dims) -> Any:
+        dims = wrap(dims)
+        if not _is_list(dims):
+            dims = wrap([dims])
+
+        raw = []
+        for i in range(len(dims)):
+            x = +dims[i]
+            y = int(x)
+            if x != y: raise ValueError(f'reshape dims must be int, got {x}')
+            if y < 0: raise ValueError(f'reshape dims cannot be negative, got {y}')
+            if y == 0: return wrap([])
+            raw.append(y)
+
+        src = self.flat
+        if len(src) == 0:
+            src = wrap([''])
+
+        i = [0]
+        def create(d):
+            if len(d) == 0:
+                r = src[i[0]]
+                i[0] += 1
+                if i[0] >= len(src):
+                    i[0] = 0
+                return r
+            else:
+                return [create(d[1:]) for _ in range(d[0])]
+        return wrap(create(raw))
+
+    @property
+    def shape(self) -> 'List':
+        res = []
+        def visitor(v, depth = 0):
+            if depth >= len(res):
+                res.append(0)
+            res[depth] = max(res[depth], len(v))
+            for x in v:
+                if _is_list(x):
+                    visitor(x, depth + 1)
+        visitor(self)
+        return wrap(res)
+    @property
+    def flat(self) -> 'List':
+        res = []
+        def visitor(v):
+            for x in v:
+                if _is_list(x):
+                    visitor(x)
+                else:
+                    res.append(x)
+        visitor(self)
+        return wrap(res)
+    @property
+    def T(self) -> 'List':
+        columns = max([0, *[len(x) if _is_list(x) else 1 for x in self]])
+        res = []
+        for column in range(columns):
+            inner = []
+            for row in self:
+                inner.append(row[column] if _is_list(row) else row)
+            res.append(inner)
+        return wrap(res)
+
+    @property
+    def csv(self):
+        res = io.StringIO()
+        writer = csv.writer(res, lineterminator = '')
+        if any(_is_list(x) for x in self):
+            for i, row in enumerate(self):
+                if i != 0: res.write('\n')
+                writer.writerow(row)
+        else:
+            writer.writerow(self)
+        return wrap(res.getvalue())
+
     def pop(self) -> Any:
         if len(self) == 0: return wrap('')
         res = wrap(list.__getitem__(self, -1))
         list.__delitem__(self, -1)
         return res
-    def all_but_first(self) -> 'List':
-        if len(self) == 0: return List()
-        return wrap(list.__getitem__(self, slice(1, len(self))))
 
     def index(self, *args, **kwargs) -> Float:
         try:
@@ -362,6 +528,20 @@ def srange(a: Any, b: Any) -> List:
     '''
     return _list_unary_op(sxrange(a, b), List)
 
+def combinations(*sources: Any) -> Sequence[Any]:
+    '''
+    Returns a list of combinations of items in the source lists.
+    With one list, this returns `[x]` for each `x` in in the list.
+    With two lists, this returns `[x, y]` for each `x` and `y` the first and second lists, respectively.
+    And so on.
+    '''
+    sources = wrap(sources)
+    if len(sources) == 0: return wrap([])
+    return wrap(list(itertools.product(*sources)))
+
+def log(value: Any, base: Any) -> Any:
+    return _list_binary_op(wrap(value), wrap(base), lambda x, y: wrap(math.log(+x, +y)))
+
 def sqrt(value: Any) -> Any:
     return _list_unary_op(wrap(value), lambda x: wrap(math.sqrt(+x)))
 def lnot(value: Any) -> Any:
@@ -381,11 +561,17 @@ def acos(value: Any) -> Any:
 def atan(value: Any) -> Any:
     return _list_unary_op(wrap(value), lambda x: wrap(math.atan(+x) * (180 / math.pi)))
 
+def atan2(y: Any, x: Any) -> Any:
+    return _list_binary_op(wrap(y), wrap(x), lambda a, b: wrap(math.atan2(+a, +b) * (180 / math.pi)))
+
 def get_ord(value: Any) -> Any:
     return _list_unary_op(wrap(value), lambda x: wrap(ord(str(x))))
 def get_chr(value: Any) -> Any:
     return _list_unary_op(wrap(value), lambda x: wrap(chr(+x)))
 
+def sign(value: Any) -> Any:
+    return _list_unary_op(wrap(value), lambda x: wrap(_single_cmp(x, wrap(0))))
+
 def identical(a: Any, b: Any) -> Any:
     a, b = wrap(a), wrap(b)
     la, lb = _is_list(a), _is_list(b)
@@ -393,6 +579,46 @@ def identical(a: Any, b: Any) -> Any:
     if la or lb: return False
     return a == b
 
+def prod(vals: Any) -> Any:
+    vals = wrap(vals)
+    if type(vals) is List: return _list_fold_op(vals, wrap(1), lambda x, y: x * y)
+    else: return vals
+
+def split(src: Any, by: Any) -> List:
+    def splitter(x, y):
+        x, y = str(x), str(y)
+        return wrap(x.split(y) if y != '' else list(x))
+    return _list_binary_op(wrap(src), wrap(by), splitter)
+def split_csv(src: Any) -> List:
+    def splitter(x):
+        raw = list(csv.reader(str(x).splitlines()))
+        return wrap(raw[0]) if len(raw) == 1 else wrap(raw)
+    return _list_unary_op(wrap(src), splitter)
+def split_words(src: Any) -> List:
+    return _list_unary_op(wrap(src), lambda x: wrap(re.split(r'\s+', str(x))))
+def split_json(src: Any) -> List:
+    return _list_unary_op(wrap(src), lambda x: wrap(json.loads(str(x))))
+
+def is_number(value: Any) -> bool:
+    if isinstance(value, bool): return False
+    try:
+        x = +wrap(value)
+        return isinstance(x, int) or (isinstance(x, float) and not math.isnan(x))
+    except:
+        return False
+def is_text(value: Any) -> bool:
+    return not isinstance(value, bool) and (isinstance(value, str) or isinstance(value, int) or isinstance(value, float)) and not is_number(value)
+def is_bool(value: Any) -> bool:
+    return isinstance(value, bool)
+def is_list(value: Any) -> bool:
+    return isinstance(value, list) or isinstance(value, tuple) or isinstance(value, dict)
+def is_sprite(value: Any) -> bool:
+    return isinstance(value, _netsblox.graphical.SpriteBase)
+def is_costume(value: Any) -> bool:
+    return isinstance(value, Image.Image)
+def is_sound(value: Any) -> bool:
+    return isinstance(value, _netsblox.sound.Sound)
+
 if __name__ == '__main__':
     assert is_wrapped(True)
     v = wrap('hello world') ; assert v is wrap(v) and isinstance(v, Str) and isinstance(v, str)
@@ -405,7 +631,7 @@ if __name__ == '__main__':
     v = wrap({1,3,2,54}) ; assert v is wrap(v) and isinstance(v, List) and isinstance(v, list)
     v = wrap({1:1,3:4,2:2,54:3}) ; assert v is wrap(v) and isinstance(v, List) and isinstance(v, list)
     assert all(isinstance(x, List) and len(x) == 2 for x in v)
-    assert v[1][0] == 3 and v[1][1] == 4 and v.last()[0] == 54 and v.last().last() == 3
+    assert v[1][0] == 3 and v[1][1] == 4 and v.last[0] == 54 and v.last.last == 3
     v = wrap([]) ; v.append(5) ; assert is_wrapped(v['0'])
     v = wrap([{'foo': 'bar'}]) ; assert isinstance(list.__getitem__(v, 0), List) and isinstance(list.__getitem__(list.__getitem__(v, 0), 0), List)
     assert list.__getitem__(list.__getitem__(v, 0), 0) == ['foo', 'bar'] and is_wrapped(list.__getitem__(list.__getitem__(list.__getitem__(v, 0), 0), 0))
@@ -495,11 +721,11 @@ if __name__ == '__main__':
     assert v[34] == '' and is_wrapped(v[34])
     assert v.index(wrap('4')) == 1 and v.index('4') == 1 and v.index(4.0) == 1 and v.index('2') == 2 and is_wrapped(v.index('2'))
     assert v.index(7) == -1 and is_wrapped(v.index(7))
-    assert wrap([]).random() == '' and is_wrapped(wrap([]).random())
-    assert wrap([]).last() == '' and is_wrapped(wrap([]).last())
-    assert wrap([3]).last() == 3 and is_wrapped(wrap([3]).last())
-    assert wrap([3, 46]).last() == 46 and is_wrapped(wrap([3, 46]).last())
-    assert wrap([3, 101, 46]).last() == 46 and is_wrapped(wrap([3, 101, 46]).last())
+    assert wrap([]).rand == '' and is_wrapped(wrap([]).rand)
+    assert wrap([]).last == '' and is_wrapped(wrap([]).last)
+    assert wrap([3]).last == 3 and is_wrapped(wrap([3]).last)
+    assert wrap([3, 46]).last == 46 and is_wrapped(wrap([3, 46]).last)
+    assert wrap([3, 101, 46]).last == 46 and is_wrapped(wrap([3, 101, 46]).last)
 
     assert identical(wrap('768') + wrap('0'), wrap('768')) and identical(wrap('768') + wrap('0'), wrap(768))
     assert not identical(wrap([1,2,3]), wrap([1,2,3]))
@@ -556,19 +782,19 @@ if __name__ == '__main__':
     vv = v[[[4,2],2,[[3,1]]]] ; assert vv == [['e', 'c'], 'c', [['d', 'b']]] and is_wrapped(vv)
 
     v = wrap([])
-    vv = v.random() ; assert vv == '' and is_wrapped(vv)
-    vv = v.all_but_first() ; assert vv == [] and is_wrapped(vv) and vv is not v and not identical(vv, v)
+    vv = v.rand ; assert vv == '' and is_wrapped(vv)
+    vv = v[1:] ; assert vv == [] and is_wrapped(vv) and vv is not v and not identical(vv, v)
     v = wrap([7])
-    vv = v.random() ; assert vv == 7 and is_wrapped(vv)
-    vv = v.all_but_first() ; assert vv == [] and is_wrapped(vv) and vv is not wrap([]).all_but_first()
+    vv = v.rand ; assert vv == 7 and is_wrapped(vv)
+    vv = v[1:] ; assert vv == [] and is_wrapped(vv) and vv is not wrap([])[1:]
     v = wrap([3, 6, 3, 8, 7])
-    vv = v.all_but_first() ; assert vv == [6, 3, 8, 7] and is_wrapped(vv) and v == [3,6,3,8,7]
-    xx = vv.pop() ; assert xx == 7 and is_wrapped(xx) and vv == [6,3,8] and vv.last() == 8 and vv.last() == 8 and is_wrapped(vv.last())
-    xx = vv.pop() ; assert xx == 8 and is_wrapped(xx) and vv == [6,3] and vv.last() == 3 and vv.last() == 3 and is_wrapped(vv.last())
-    xx = vv.pop() ; assert xx == 3 and is_wrapped(xx) and vv == [6] and vv.last() == 6 and vv.last() == 6 and is_wrapped(vv.last())
-    xx = vv.pop() ; assert xx == 6 and is_wrapped(xx) and vv == [] and vv.last() == '' and vv.last() == '' and is_wrapped(vv.last())
-    xx = vv.pop() ; assert xx == '' and is_wrapped(xx) and vv == [] and vv.last() == '' and vv.last() == '' and is_wrapped(vv.last())
-    xx = vv.pop() ; assert xx == '' and is_wrapped(xx) and vv == [] and vv.last() == '' and vv.last() == '' and is_wrapped(vv.last())
+    vv = v[1:] ; assert vv == [6, 3, 8, 7] and is_wrapped(vv) and v == [3,6,3,8,7]
+    xx = vv.pop() ; assert xx == 7 and is_wrapped(xx) and vv == [6,3,8] and vv.last == 8 and vv.last == 8 and is_wrapped(vv.last)
+    xx = vv.pop() ; assert xx == 8 and is_wrapped(xx) and vv == [6,3] and vv.last == 3 and vv.last == 3 and is_wrapped(vv.last)
+    xx = vv.pop() ; assert xx == 3 and is_wrapped(xx) and vv == [6] and vv.last == 6 and vv.last == 6 and is_wrapped(vv.last)
+    xx = vv.pop() ; assert xx == 6 and is_wrapped(xx) and vv == [] and vv.last == '' and vv.last == '' and is_wrapped(vv.last)
+    xx = vv.pop() ; assert xx == '' and is_wrapped(xx) and vv == [] and vv.last == '' and vv.last == '' and is_wrapped(vv.last)
+    xx = vv.pop() ; assert xx == '' and is_wrapped(xx) and vv == [] and vv.last == '' and vv.last == '' and is_wrapped(vv.last)
     v = wrap([3, 6, 3, 8, 7])
     assert v == [3,6,3,8,7] and len(v) == 5
     v.clear()
@@ -592,4 +818,234 @@ if __name__ == '__main__':
     v = +wrap('-Infinity') ; assert v < 0 and math.isinf(v)
     v = +wrap('NaN') ; assert math.isnan(v)
 
+    assert prod([4, 2, 6]) == 48 and is_wrapped(prod([4, 2, 6]))
+    assert prod([4]) == 4 and is_wrapped(prod([4]))
+    assert prod([]) == 1 and is_wrapped(prod([]))
+    assert prod([[4, 2, 6], [2, 7, 3]]) == [8, 14, 18] and is_wrapped(prod([[4, 2, 6], [2, 7, 3]]))
+
+    assert (wrap('7') > 4) == True and (wrap('7') > '4') == True and (wrap('7') >= 4) == True and (wrap('7') >= '4') == True
+    assert (wrap('7') < 12) == True and (wrap('7') < '12') == True and (wrap('7') <= 12) == True and (wrap('7') <= '12') == True
+    assert (wrap('7') == 7) == True and (wrap('7') == '7') == True and (wrap('7') != 12) == True and (wrap('7') != '12') == True
+    assert (wrap('7') < 4) == False and (wrap('7') < '4') == False and (wrap('7') <= 4) == False and (wrap('7') <= '4') == False
+    assert (wrap('7') > 12) == False and (wrap('7') > '12') == False and (wrap('7') >= 12) == False and (wrap('7') >= '12') == False
+    assert (wrap('7') == 12) == False and (wrap('7') == '12') == False and (wrap('7') != 7) == False and (wrap('7') != '7') == False
+
+    assert (wrap(7) > 4) == True and (wrap(7) > '4') == True and (wrap(7) >= 4) == True and (wrap(7) >= '4') == True
+    assert (wrap(7) < 12) == True and (wrap(7) < '12') == True and (wrap(7) <= 12) == True and (wrap(7) <= '12') == True
+    assert (wrap(7) == 7) == True and (wrap(7) == '7') == True and (wrap(7) != 12) == True and (wrap(7) != '12') == True
+    assert (wrap(7) < 4) == False and (wrap(7) < '4') == False and (wrap(7) <= 4) == False and (wrap(7) <= '4') == False
+    assert (wrap(7) > 12) == False and (wrap(7) > '12') == False and (wrap(7) >= 12) == False and (wrap(7) >= '12') == False
+    assert (wrap(7) == 12) == False and (wrap(7) == '12') == False and (wrap(7) != 7) == False and (wrap(7) != '7') == False
+
+    assert (wrap('7') < [1, 5, 7, 9, 10]) == [False, False, False, True, True]
+    assert (wrap('7') <= [1, 5, 7, 9, 10]) == [False, False, True, True, True]
+    assert (wrap('7') > [1, 5, 7, 9, 10]) == [True, True, False, False, False]
+    assert (wrap('7') >= [1, 5, 7, 9, 10]) == [True, True, True, False, False]
+    assert (wrap('7') >= [1, 5, 7, 9, 10]) != [True, True, False, False, False]
+
+    assert (wrap(7) < ['1', 5, 7, 9, 10]) == [False, False, False, True, True]
+    assert (wrap(7) <= [1, '5', 7, 9, 10]) == [False, False, True, True, True]
+    assert (wrap(7) > [1, 5, '7', 9, 10]) == [True, True, False, False, False]
+    assert (wrap(7) >= [1, 5, 7, '9', 10]) == [True, True, True, False, False]
+    assert (wrap(7) >= [1, 5, 7, 9, '10']) != [True, True, False, False, False]
+
+    assert (wrap(['1', 5, 7, 9, 10]) > 7) == [False, False, False, True, True]
+    assert (wrap([1, '5', 7, 9, 10]) >= 7) == [False, False, True, True, True]
+    assert (wrap([1, 5, '7', 9, 10]) < 7) == [True, True, False, False, False]
+    assert (wrap([1, 5, 7, '9', 10]) <= 7) == [True, True, True, False, False]
+    assert (wrap([1, 5, 7, 9, '10']) <= 7) != [True, True, False, False, False]
+
+    assert (wrap(['1', 5, 7, 9, 10]) > '7') == [False, False, False, True, True]
+    assert (wrap([1, '5', 7, 9, 10]) >= '7') == [False, False, True, True, True]
+    assert (wrap([1, 5, '7', 9, 10]) < '7') == [True, True, False, False, False]
+    assert (wrap([1, 5, 7, '9', 10]) <= '7') == [True, True, True, False, False]
+    assert (wrap([1, 5, 7, 9, '10']) <= '7') != [True, True, False, False, False]
+
+    assert sign(6) == 1 and is_wrapped(sign(6))
+    assert sign('-7') == -1 and is_wrapped(sign(-7))
+    assert sign(0) == 0 and is_wrapped(sign(0))
+    assert sign(['5', -2, '0', 3]) == [1, -1, 0, 1] and is_wrapped(sign([5, -2, 0, 3]))
+
+    assert is_wrapped(atan2(1, '2')) and abs(atan2('1', 2) - 26.5650511) < 0.00001
+    assert is_wrapped(log(10, '2')) and abs(log('10', 2) - 3.321928) < 0.000001
+
+    assert split('hello world', ' ') == ['hello', 'world']
+    assert split('hello  world', ' ') == ['hello', '', 'world']
+    assert split(['hello  world', 'again'], ' ') == [['hello', '', 'world'], ['again']]
+    assert split('hello world', '') == ['h', 'e', 'l', 'l', 'o', ' ', 'w', 'o', 'r', 'l', 'd']
+    assert split(['hello world', 'more'], '') == [['h', 'e', 'l', 'l', 'o', ' ', 'w', 'o', 'r', 'l', 'd'], ['m', 'o', 'r', 'e']]
+    assert split('hello\nworld', '\n') == ['hello', 'world']
+
+    assert split_words('hello world') == ['hello', 'world']
+    assert split_words('hello  world') == ['hello', 'world']
+    assert split_words(['hello world', 'hello    \t \n again']) == [['hello', 'world'], ['hello', 'again']]
+
+    assert split_json('["this", 12, true, "is a", "test"]') == ['this', 12, True, 'is a', 'test']
+    assert split_json('{"a": 12, "b": ["x", "y", "z"]}') == [['a', 12], ['b', ['x', 'y', 'z']]]
+    assert split_json('{"a": 12, "b": ["x", "y", "z"]}') != [['a', 12], ['b', ['x', 'y', 'zz']]]
+
+    assert split_csv('"this, is some text",12,true,","') == ['this, is some text', '12', 'true', ',']
+    assert split_csv('"this, is some text",12,true,","\n12\n\ntrue,') == [['this, is some text', '12', 'true', ','], ['12'], [], ['true', '']]
+
+    assert wrap('z') == wrap('Z') and wrap('aBc') == 'Abc' and 'abC' == wrap('ABC')
+    assert wrap({}) == [] and wrap({'a': 5}) == [['a', 5]]
+
+    assert is_number('5') and is_number('-5.8') and is_number(wrap('5')) and is_number(5) and is_number(5.6) and is_number(wrap(5)) and not is_number('hello') and not is_number(wrap('hello')) and not is_number(True) and not is_number([]) and not is_number({}) and not is_number('{}') and not is_number(math.nan)
+    assert is_text('hello') and is_text(wrap('hello')) and not is_text(5) and not is_text('5') and not is_text([]) and not is_text(True) and not is_text({})
+    assert is_bool(True) and is_bool(False) and is_bool(wrap(True)) and is_bool(wrap(False)) and not is_bool(12) and not is_bool('56') and not is_bool('hello') and not is_bool([]) and not is_bool({})
+    assert is_list([]) and is_list({}) and is_list(wrap([])) and is_list(wrap({})) and not is_list('hello') and not is_list(wrap('hello')) and not is_list(45) and not is_list(wrap(45)) and not is_list(True) and not is_list(False) and is_list(())
+
+    v = wrap([1, 5, 3, 8, 6, 4])
+    assert v[3] == 8 and v[[2, 4]] == [3, 6] and v[2,[4],2,[[1]]] == [3, [6], 3, [[5]]]
+
+    v = wrap([1, 4, 3])
+    vv = wrap([*v])
+    assert v == vv and v is not vv
+    v = v + 1
+    assert v == [2, 5, 4] and vv == [1, 4, 3]
+    v += 1
+    assert v == [3, 6, 5] and vv == [1, 4, 3]
+    v *= 2
+    assert v == [6, 12, 10] and vv == [1, 4, 3]
+    v /= 2
+    assert v == [3, 6, 5] and vv == [1, 4, 3]
+    v -= vv
+    assert v == [2, 2, 2] and vv == [1, 4, 3]
+    v = wrap([1, 2, 3])
+    v **= 2
+    assert v == [1, 4, 9]
+
+    v = wrap('12')
+    v += 5
+    assert v == 17
+    v += '2'
+    assert v == 19
+
+    v = wrap(12)
+    v += 5
+    assert v == 17
+    v += '2'
+    assert v == 19
+
+    v = wrap('45')
+    v += '12'
+    assert v == 57
+
+    v = wrap([1, 3, 4, 2])
+    assert v[1:] == [3, 4, 2] and is_wrapped(v[1:])
+    assert v.last == 2
+    assert wrap([3, 3, 3]).rand == 3
+    assert v.index(3) == 1
+    assert 3 in v and wrap(3) in v and 5 not in v and wrap(5) not in v
+    assert len(v) == 4
+    assert v[::-1] == [2, 4, 3, 1] and is_wrapped(v[::-1])
+    assert v.shape == [4] and is_wrapped(v.shape)
+    assert wrap([4, 3, [], 4]).shape == [4, 0]
+    assert wrap([4, 3, [], 4, [6]]).shape == [5, 1]
+    assert wrap([4, 3, [[]], 4, [6], [[[]]]]).shape == [6, 1, 1, 0]
+    assert wrap([]).shape == [0]
+    assert wrap([[]]).shape == [1, 0]
+
+    assert wrap([]).flat == [] and is_wrapped(wrap([]).flat)
+    assert wrap([4, 3, 2]).flat == [4, 3, 2] and is_wrapped(wrap([4, 3, 2]).flat)
+    assert wrap([4, [[5]], [], [6, [4], 3, 1], 0]).flat == [4, 5, 6, 4, 3, 1, 0]
+
+    assert wrap([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).T == [[1, 4, 7], [2, 5, 8], [3, 6, 9]]
+    assert wrap([[1, 2], [4, 5, 6], [7, 8]]).T == [[1, 4, 7], [2, 5, 8], ['', 6, '']]
+    assert wrap([[1, 2, 3], [4, 5], [7, 8, 9]]).T == [[1, 4, 7], [2, 5, 8], [3, '', 9]]
+    assert wrap([[1, 2], [4, 5], [7, 8, 9]]).T == [[1, 4, 7], [2, 5, 8], ['', '', 9]]
+    assert wrap([[1, 2, 3], [4, 5], [7, 8]]).T == [[1, 4, 7], [2, 5, 8], [3, '', '']]
+    assert wrap([]).T == [] and wrap([[]]).T == [] and wrap([[]]).shape == [1, 0] and wrap([[]]).T.shape == [0]
+    assert wrap([1, 2, 3, 4, 5]).T == [[1, 2, 3, 4, 5]]
+    assert wrap([[1,2,3],[4,5],[6]]).T == [[1,4,6],[2,5,''],[3,'','']]
+    assert wrap([[1,2,3],[4,5],[6]]).T != [[1,4,6],[2,5],[3]]
+
+    assert wrap([1, 2, 3]).csv == '1,2,3'
+    assert wrap([[' ', 2, ' ']]).csv == ' ,2, '
+    assert wrap([['some text', 2, 3],[5, 4, 6]]).csv == 'some text,2,3\n5,4,6'
+    assert wrap([[1, 2, 3],[5, ',', 6]]).csv == '1,2,3\n5,",",6'
+
+    assert wrap([1, 2, 3])[::-1] == [3, 2, 1] and wrap([4])[::-1] == [4] and wrap([])[::-1] == []
+
+    assert '\n'.join(str(x) for x in [12, 4, 'hello', 'more', 43]) == '12\n4\nhello\nmore\n43'
+
+    v = wrap([])
+    v.append(12)
+    assert is_wrapped(list.__getitem__(v, 0))
+    assert is_wrapped(v[0] is list.__getitem__(v, 0))
+    v[0] = 3
+    assert is_wrapped(list.__getitem__(v, 0))
+    assert is_wrapped(v[0] is list.__getitem__(v, 0))
+    v.append('test')
+    assert v == [3, 'test']
+    v.pop()
+    assert v == [3]
+
+    v = wrap([11, 34, 54, 21])
+    del v[2]
+    assert v == [11, 34, 21]
+    vv = v
+    assert v is vv
+    v.clear()
+    assert v is vv and v == [] and vv == []
+    v.insert(0, 'test')
+    assert v == ['test']
+    assert is_wrapped(list.__getitem__(v, 0)) and list.__getitem__(v, 0) is v[0] and v[0] == 'test'
+    v.insert(0, 'again')
+    assert v == ['again', 'test']
+    assert is_wrapped(list.__getitem__(v, 0)) and list.__getitem__(v, 0) is v[0] and v[0] == 'again'
+
+    v = wrap([1, 1, 1])
+    v.insert_rand(1)
+    assert v == [1, 1, 1, 1] and all(is_wrapped(x) for x in list.__iter__(v))
+    v.clear()
+    v.insert_rand(12)
+    assert v == [12] and all(is_wrapped(x) for x in list.__iter__(v))
+    v.rand = 5
+    assert v == [5] and all(is_wrapped(x) for x in list.__iter__(v))
+    v.rand *= 3
+    assert v == [15] and all(is_wrapped(x) for x in list.__iter__(v))
+    v = wrap([1, 4, 3])
+    assert v.last == 3 and is_wrapped(v.last)
+    v.last = 7
+    assert v == [1, 4, 7] and v.last == 7 and all(is_wrapped(x) for x in list.__iter__(v))
+    v.last += 43
+    assert v == [1, 4, 50] and v.last == 50 and all(is_wrapped(x) for x in list.__iter__(v))
+
+    assert wrap([1, 4, 3, 6]).reshaped(2) == [1, 4] and is_wrapped(wrap([1, 4, 3, 6]).reshaped(2))
+    assert wrap([1, 4, 5, 2]).reshaped(10) == [1, 4, 5, 2, 1, 4, 5, 2, 1, 4] and is_wrapped(wrap([1, 4, 5, 2]).reshaped(10))
+    assert wrap([1, 5]).reshaped([7]) == [1, 5, 1, 5, 1, 5, 1] and is_wrapped(wrap([1, 5]).reshaped([7]))
+    assert wrap([1, 7, 2]).reshaped([7, 2]) == [[1, 7], [2, 1], [7, 2], [1, 7], [2, 1], [7, 2], [1, 7]] and is_wrapped(wrap([1, 7, 2]).reshaped([7, 2]))
+    assert wrap([1, 7]).reshaped([]) == 1 and is_wrapped(wrap([1, 7]).reshaped([]))
+    assert wrap([]).reshaped([3, 2]) == [['', ''], ['', ''], ['', '']] and is_wrapped(wrap([]).reshaped([3, 2]))
+    assert wrap([1, 7, 2]).reshaped([1, 1, 7, 2]) == [[[[1, 7], [2, 1], [7, 2], [1, 7], [2, 1], [7, 2], [1, 7]]]] and is_wrapped(wrap([1, 7, 2]).reshaped([1, 1, 7, 2]))
+    assert wrap([1, 7, 4]).reshaped([1, 6, 3, 0, 2, 4]) == [] and is_wrapped(wrap([1, 7, 4]).reshaped([1, 6, 3, 0, 2, 4]))
+    assert wrap([]).reshaped([2, 2]) == [['', ''], ['', '']] and is_wrapped(wrap([]).reshaped([2, 2]))
+    assert wrap([5, 6]).reshaped([]) == 5 and is_wrapped(wrap([5, 6]).reshaped([]))
+    assert wrap([5, 6]).reshaped([2, 2, 0, 2]) == []
+
+    v = wrap([[['test']]])
+    assert is_wrapped(list.__getitem__(v, 0)) and is_wrapped(list.__getitem__(list.__getitem__(v, 0), 0)) and is_wrapped(list.__getitem__(list.__getitem__(list.__getitem__(v, 0), 0), 0))
+
+    assert combinations() == [] and is_wrapped(combinations())
+    assert combinations([1, 2]) == [[1], [2]] and is_wrapped(combinations([1, 2]))
+    assert combinations([1, 2], [3, 4]) == [[1, 3], [1, 4], [2, 3], [2, 4]] and is_wrapped(combinations([1, 2], [3, 4]))
+    assert combinations([1, 2], [], [3, 4]) == [] and is_wrapped(combinations([1, 2], [3, 4]))
+
+    v = wrap('hello world')
+    assert v[0] == 'h' and is_wrapped(v[0]) and v['0'] == 'h' and is_wrapped(v['0'])
+    assert v[4] == 'o' and is_wrapped(v[4]) and v['4'] == 'o' and is_wrapped(v['4'])
+
+    assert wrap('').last == '' and is_wrapped(wrap('').last)
+    assert wrap('begin').last == 'n' and is_wrapped(wrap('begin').last)
+    assert wrap('also').last == 'o' and is_wrapped(wrap('also').last)
+
+    assert wrap('').rand == '' and is_wrapped(wrap('').rand)
+    assert wrap('tttt').rand == 't' and is_wrapped(wrap('tttt').rand)
+
+    assert wrap([]).last == '' and is_wrapped(wrap([]).last)
+    assert wrap([]).rand == '' and is_wrapped(wrap([]).rand)
+
+    assert math.isnan(+wrap('{}'))
+
     print('passed all snap wrapper tests')
